@@ -1,7 +1,7 @@
 from coba.contexts     import CobaContext, NullLogger
 from coba.environments import Environments, ArffSource, CsvSource, LibSvmSource, ManikSource
-from coba.environments import SimulatedInteraction, LoggedInteraction, GroundedInteraction
-from coba.experiments  import Experiment, Result
+from coba.environments import SimulatedInteraction, LoggedInteraction, GroundedInteraction, LambdaSimulation
+from coba.experiments  import Experiment, Result, SimpleEvaluation, SimpleLearnerInfo, SimpleEnvironmentInfo
 
 from coba.random import CobaRandom
 
@@ -11,6 +11,8 @@ from coba.learners.vowpal import VowpalLearner, VowpalEpsilonLearner, VowpalSoft
 from coba.learners.vowpal import VowpalCoverLearner, VowpalRegcbLearner, VowpalSquarecbLearner, VowpalOffPolicyLearner
 from coba.learners.linucb import LinUCBLearner
 
+from coba.exceptions import CobaException
+
 from coba.backports import version, PackageNotFoundError
 
 try:
@@ -20,6 +22,7 @@ except PackageNotFoundError: #pragma: no cover
     __version__ = "0.0.0"
 
 __all__ = [
+    "CobaException",
     "CobaContext",
     "Environments",
     "Experiment",
@@ -42,11 +45,15 @@ __all__ = [
     'OnlineOnPolicyEval',
     'CobaRandom',
     'NullLogger',
-    'ArffSource', 
+    'ArffSource',
     'CsvSource',
     'LibSvmSource',
     'ManikSource',
-    'SimulatedInteraction', 
-    'LoggedInteraction', 
+    'SimulatedInteraction',
+    'LoggedInteraction',
     'GroundedInteraction',
+    'SimpleEvaluation', 
+    'SimpleLearnerInfo', 
+    'SimpleEnvironmentInfo',
+    'LambdaSimulation'
 ]
