@@ -13,7 +13,7 @@ from itertools import islice, chain, tee, compress, repeat
 from typing import Optional, Sequence, Union, Iterable, Any, Tuple, Callable, Mapping
 from coba.backports import Literal
 
-from coba            import pipes, primitives
+from coba            import pipes, primitives, CobaContext
 from coba.random     import CobaRandom
 from coba.exceptions import CobaException
 from coba.statistics import iqr
@@ -32,6 +32,10 @@ class Take(pipes.Take, EnvironmentFilter):
     """Take a fixed number of interactions from an Environment."""
     pass
 
+class Slice(pipes.Slice, EnvironmentFilter):
+    """Take a slice of interactions from an Environment."""
+    pass
+
 class Shuffle(pipes.Shuffle, EnvironmentFilter):
     """Shuffle a sequence of Interactions in an Environment."""
     pass
@@ -428,29 +432,36 @@ class Sparse(EnvironmentFilter):
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
+        first,interactions = peek_first(interactions)
+
+        context_has_headers = 'context' in first and hasattr(first['context']   ,'headers')
+        actions_has_headers = 'actions' in first and hasattr(first['actions'][0],'headers')
+        action_has_headers  = 'action'  in first and hasattr(first['action' ]   ,'headers')
+
         for interaction in interactions:
 
-            new = interaction.copy()            
+            new = interaction.copy()
 
             if self._context:
-                new['context'] = self._make_sparse(new['context'])
+                new['context'] = self._make_sparse(new['context'], context_has_headers, 'context')
 
             if self._action and 'actions' in new:
-                new['actions'] = list(map(self._make_sparse,new['actions']))
+                new['actions'] = list(map(self._make_sparse,new['actions'],repeat(actions_has_headers),repeat('action')))
 
             if self._action and 'action' in new:
-                new['action'] = self._make_sparse(new['action'])
+                new['action'] = self._make_sparse(new['action'],action_has_headers, 'action')
 
             yield new
 
-    def _make_sparse(self, value) -> Optional[dict]:
+    def _make_sparse(self, value, has_headers:bool, default_header:str) -> Optional[dict]:
         if value is None:
             return value
-        if isinstance(value,primitives.Dense):
-            return dict(enumerate(value))
+        if isinstance(value,primitives.Dense): 
+            value_list = list(value)
+            return {k:value_list[i] for k,i in value.headers.items() if i < len(value_list)} if has_headers else dict(enumerate(value))
         if isinstance(value,primitives.Sparse):
             return value
-        return {0:value}
+        return {default_header:value}
 
 class Cycle(EnvironmentFilter):
     """Cycle all rewards associated with actions by one place.
@@ -481,23 +492,26 @@ class Cycle(EnvironmentFilter):
         first, with_cycle_interactions = peek_first(with_cycle_interactions)
 
         if with_cycle_interactions:
-            action_set          = set(first['actions'])
-            n_actions           = len(action_set)
-            featureless_actions = [tuple([0]*n+[1]+[0]*(n_actions-n-1)) for n in range(n_actions)]
 
-            is_discrete = first['actions'] and 0 < len(first['actions']) and len(first['actions']) < float('inf')
+            action_set = set(first['actions'])
+            n_actions  = len(action_set)
 
-            if not is_discrete or len(set(action_set) & set(featureless_actions)) != len(action_set):
+            onehot_actions  = {tuple([0]*n+[1]+[0]*(n_actions-n-1)) for n in range(n_actions)}
+            is_discrete     = 0 < n_actions and n_actions < float('inf')
+            is_onehots      = action_set == onehot_actions
+            is_categoricals = all([isinstance(a,str) for a in action_set])
+
+            if not is_discrete or (not is_onehots and not is_categoricals):
                 warnings.warn("Cycle only works for discrete environments without action features. It will be ignored in this case.")
                 yield from with_cycle_interactions
             else:
                 for interaction in with_cycle_interactions:
                     rewards = deque(interaction['rewards'])
                     rewards.rotate(1)
-                    
+
                     new = interaction.copy()
                     new['rewards'] = SequenceReward(list(rewards))
-                    
+
                     yield new
 
 class Flatten(EnvironmentFilter):
@@ -908,18 +922,25 @@ class Unbatch(EnvironmentFilter):
 
         first, interactions = peek_first(interactions)
         if not interactions: return []
-        is_batched = isinstance(first[list(first.keys())[0]], primitives.Batch)
+        
+        batched_keys = [k for k,v in first.items() if isinstance(v,primitives.Batch) ]
 
-        if not is_batched:
+        if not batched_keys:
             yield from interactions
         else:
-            yield from self._unbatch(interactions)
+            yield from self._unbatch(interactions, batched_keys)
 
-    def _unbatch(self, interactions: Iterable[Interaction]):
+    def _unbatch(self, interactions: Iterable[Interaction], batched_keys:Sequence[str]):
         for interaction in interactions:
-            batch_size = len(interaction[list(interaction.keys())[0]])
+            batch_size = len(interaction[batched_keys[0]])
             for i in range(batch_size):
-                yield { k: interaction[k][i] for k in interaction }
+                new = {}
+                for k in interaction:
+                    try:
+                        new[k] = interaction[k][i]
+                    except:
+                        new[k] = interaction[k]
+                yield new
 
 class BatchSafe(EnvironmentFilter):
 
@@ -954,20 +975,20 @@ class Finalize(EnvironmentFilter):
         first_has_actions = 'actions' in first and first['actions']
         first_has_action  = 'action'  in first
 
-        context_materialized = not first_has_context or isinstance(first['context']   ,(list,tuple,dict))
-        actions_materialized = not first_has_actions or isinstance(first['actions'][0],(list,tuple,dict))
-        action_materialized  = not first_has_action  or isinstance(first['action' ]   ,(list,tuple,dict))
-
         if first_has_context:
             is_dense_context  = isinstance(first['context'],primitives.Dense)
             is_sparse_context = isinstance(first['context'],primitives.Sparse)
         if first_has_actions:
-            is_dense_action  = isinstance(first['actions'][0],primitives.Dense)
-            is_sparse_action = isinstance(first['actions'][0],primitives.Sparse)
-        elif first_has_action:
+            is_dense_actions  = isinstance(first['actions'][0],primitives.Dense)
+            is_sparse_actions = isinstance(first['actions'][0],primitives.Sparse)
+        if first_has_action:
             is_dense_action  = isinstance(first['action'],primitives.Dense)
             is_sparse_action = isinstance(first['action'],primitives.Sparse)
 
+        context_materialized = not first_has_context or (not is_dense_context and not is_sparse_context or isinstance(first['context']   ,(list,tuple,dict)))
+        actions_materialized = not first_has_actions or (not is_dense_actions and not is_sparse_actions or isinstance(first['actions'][0],(list,tuple,dict)))
+        action_materialized  = not first_has_action  or (not is_dense_action  and not is_sparse_action  or isinstance(first['action']    ,(list,tuple,dict)))
+
         for interaction in Repr("onehot","onehot").filter(interactions):
 
             new = interaction.copy()
@@ -975,9 +996,9 @@ class Finalize(EnvironmentFilter):
             if not context_materialized and is_dense_context:
                 new['context'] = list(new['context'])
             elif not context_materialized and is_sparse_context :
-                    new['context'] = new['context'].copy()
+                new['context'] = new['context'].copy()
 
-            if not actions_materialized and is_dense_action:
+            if not actions_materialized and is_dense_actions:
                 new['actions'] = list(map(list,new['actions']))
             elif not actions_materialized and is_sparse_action:
                 new['actions'] = list(map(methodcaller('copy'),new['actions']))
@@ -1001,19 +1022,41 @@ class Logged(EnvironmentFilter):
 
     @property
     def params(self) -> Mapping[str, Any]:
-        return {"learner": self._learner.params}
+        return {"learner": SafeLearner(self._learner).params, "logged":True}
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
+        first, interactions = peek_first(interactions)
+
+        if not interactions: return []
+
+        if 'context' not in first or 'actions' not in first or 'rewards' not in first:
+            raise CobaException("We were unable to create a logged representation of the interaction.")
+
         learner = SafeLearner(copy.deepcopy(self._learner))
 
-        first, interactions = peek_first(interactions)
+        #In theory all of the code below could be replaced by:
+
+            # I1,I2 = tee(interactions,2)
+            # for interaction, log in zip(I1, SimpleEvaluation(record=['action','reward','probability']).filter(I2)):
+            #     out = interaction.copy()
+
+            #     out['action']      = log['action']
+            #     out['reward']      = log['reward']
+            #     out['probability'] = log['probability']
+
+            #     yield out
+
+        #The reason it hasn't is because referencing SimpleEvalution creates a recursive dependency 
 
         predict = learner.predict
         learn   = learner.learn
 
-        if 'context' not in first or 'actions' not in first or 'rewards' not in first:
-            raise CobaException("We were unable to create a logged representation of the interaction.")
+        learning_info = CobaContext.learning_info
+
+        def indexify(action,actions,discrete,batched):
+            if not discrete: return action
+            return [ A.index(a) for a,A in zip(action,actions) ] if batched else actions.index(action)
 
         for interaction in interactions:
 
@@ -1021,9 +1064,12 @@ class Logged(EnvironmentFilter):
             actions = interaction['actions']
             rewards = interaction['rewards']
 
+            batched = isinstance(context, primitives.Batch)
+            discrete = len(interaction.get('actions',[])) > 0
+
             action,prob,info = predict(context, actions)
 
-            reward = rewards.eval(action)
+            reward = rewards.eval(indexify(action,actions,discrete,batched))
 
             learn(context, actions, action, reward, prob, **info)
 
@@ -1033,6 +1079,10 @@ class Logged(EnvironmentFilter):
             new['probability'] = prob
             new['reward']      = reward
 
+            if learning_info:
+                new.update(learning_info)
+                learning_info.clear()
+
             yield new
 
 class Mutable(EnvironmentFilter):
