@@ -5,13 +5,14 @@ import collections.abc
 from copy import copy
 from pathlib import Path
 from numbers import Number
-from operator import truediv, sub, gt
+from operator import truediv, sub, gt, itemgetter
 from abc import abstractmethod
 from itertools import chain, repeat, accumulate, groupby, count
 from typing import Any, Dict, List, Set, Tuple, Optional, Sequence, Hashable, Iterable, Iterator, Union, Type, Callable, NamedTuple
 from coba.backports import Literal
 
-from coba.statistics import Mean, StandardDeviation, StandardErrorOfMean, BootstrapConfidenceInterval, BinomialConfidenceInterval, PointAndInterval
+from coba.environments import Environment
+from coba.statistics import mean, stdev, StandardErrorOfMean, BootstrapConfidenceInterval, BinomialConfidenceInterval, PointAndInterval
 from coba.contexts import CobaContext
 from coba.exceptions import CobaException
 from coba.utilities import PackageChecker
@@ -443,6 +444,7 @@ class Points(NamedTuple):
     alpha: float            = 1
     label: Optional[str]    = None
     style: Literal['-','.'] = "-"
+    zorder:int              = 1
 
 class Plotter:
     @abstractmethod
@@ -456,6 +458,8 @@ class Plotter:
         ylim: Tuple[Optional[Number],Optional[Number]],
         xticks: bool,
         yticks: bool,
+        xrotation: Optional[float],
+        yrotation: Optional[float],
         out: Union[None,Literal['screen'],str]) -> None:
         pass
 
@@ -511,7 +515,7 @@ class MatplotlibPlotter(Plotter):
             any_label = False
             num_coalesce = lambda x1,x2: x1 if isinstance(x1,(int,float)) else x2
 
-            for X, Y, XE, YE, c, a, l, fmt in lines:
+            for X, Y, XE, YE, c, a, l, fmt,z in lines:
 
                 if l: any_label = True
 
@@ -529,12 +533,13 @@ class MatplotlibPlotter(Plotter):
 
                 if X and Y:
                     if all(map(not_err_bar,[XE,YE])):
-                        ax.plot(X, Y, fmt,  color=c, alpha=a, label=l)
+                        ax.plot(X, Y, fmt,  color=c, alpha=a, label=l,zorder=z)
                     else:
                         XE = None if not_err_bar(XE) else list(zip(*XE)) if isinstance(XE[0],tuple) else XE
                         YE = None if not_err_bar(YE) else list(zip(*YE)) if isinstance(YE[0],tuple) else YE
                         error_every = max(int(len(X)*0.05),1) if fmt == "-" else 1
-                        ax.errorbar(X, Y, YE, XE, fmt, elinewidth=0.5, errorevery=error_every, color=c, alpha=a, label=l)
+                        elinewidth = 0.5 if 'elinewidth' not in CobaContext.store else CobaContext.store['elinewidth']
+                        ax.errorbar(X, Y, YE, XE, fmt, elinewidth=elinewidth, errorevery=error_every, color=c, alpha=a, label=l,zorder=z)
 
             if xrotation is not None:
                 plt.xticks(rotation=xrotation)
@@ -591,14 +596,17 @@ class FilterPlottingData:
 
         learner_count   = len(set([row['learner_id'] for row in rows]))
         environ_counts  = collections.Counter([row['environment_id'] for row in rows])
-        environ_lengths = set([len(row[y]) for row in rows])
+        try:
+            environ_lengths = set([len(row[y]) for row in rows])
+        except: #pragma: no cover
+            raise CobaException(f"{y} is not available in the environment. Plotting has been stopped.")
         min_env_length  = min(environ_lengths) 
 
         if max(environ_counts.values()) != learner_count:
             raise CobaException("This result does not contain an environment which has been finished for every learner. Plotting has been stopped.")
 
         if min(environ_counts.values()) != learner_count:
-            CobaContext.logger.log("This result contains environments not present for all learners. Environments not present for all learners have been excluded. To supress this warning in the future call <result>.filter_fin() before plotting.") 
+            CobaContext.logger.log("Environments not present for all learners have been excluded. To supress this call filter_fin() before plotting.")
 
         if len(environ_lengths) > 1 and x == ['index']:
             CobaContext.logger.log("This result contains environments of different lengths. The plot only includes interactions up to the shortest environment. To supress this warning in the future call <result>.filter_fin(n_interactions) before plotting.")
@@ -618,7 +626,7 @@ class SmoothPlottingData:
 
 class ContrastPlottingData:
 
-    def filter(self, rows:Sequence[Dict[str,Any]], y:str, mode:str, learner_id1:int) -> Sequence[Dict[str,Any]]:
+    def filter(self, rows:Sequence[Dict[str,Any]], y:str, mode:Union[Literal["diff","prob",'scat'],Callable[[float,float],float]], learner_id1:int) -> Sequence[Dict[str,Any]]:
 
         sort_key  = lambda row: (row['environment_id'], 0 if row['learner_id']==learner_id1 else 1)
         group_key = lambda row: row['environment_id']
@@ -627,8 +635,10 @@ class ContrastPlottingData:
             contraster = lambda Y1,Y2: list(map(int,map(gt,Y1,Y2))) 
         elif mode == "diff": 
             contraster = lambda Y1,Y2: list(map(sub,Y1,Y2))
-        else:
+        elif mode =="scat":
             contraster = lambda Y1,Y2: list(zip(Y1,Y2))
+        else:
+            contraster = lambda Y1,Y2: list(map(mode,Y1,Y2))
 
         contrast = lambda env_id, pair: {'environment_id': env_id, y: contraster(pair[0][y],pair[1][y]) }
 
@@ -644,20 +654,17 @@ class TransformToXYE:
         err:Union[str,None,PointAndInterval]) -> Sequence[Tuple[Any,float,Union[None,float,Tuple[float,float]]]]:
 
         if err == 'sd':
-            avg = Mean()
-            std = StandardDeviation()
-            YE = lambda z: (avg.calculate(z), std.calculate(z))
+            YE = lambda z: (mean(z), stdev(z))
         elif err == 'se':
-            YE = StandardErrorOfMean().calculate
+            YE = StandardErrorOfMean(1.96).calculate #z-score for .975 area to the left and right
         elif err == "bs":
-            YE = BootstrapConfidenceInterval(.95, Mean().calculate).calculate
+            YE = BootstrapConfidenceInterval(.95, mean).calculate
         elif err == "bi":
             YE = BinomialConfidenceInterval('wilson').calculate
         elif isinstance(err,PointAndInterval):
             YE = err.calculate
         else:
-            avg = Mean()
-            YE = lambda z: (avg.calculate(z) if len(z) > 1 else z[0], None)
+            YE = lambda z: (mean(z) if len(z) > 1 else z[0], None)
 
         iters = [ iter(row[y]) for row in rows ]
         first_val = next(iters[0])
@@ -675,9 +682,9 @@ class TransformToXYE:
                 points = [YE(z1)+YE(z2) for _,z in zip(X,Z) for z1,z2 in [tuple(zip(*z))]]
         else:
             XZ = collections.defaultdict(list)
-            make_x = lambda env: env[x[0]] if len(x) == 1 else tuple(env[k] for k in x)
+            make_x = lambda eid: eid if list(x) == ['environment_id'] else envs[eid].get(x[0]) if len(x) == 1 else tuple(envs[eid].get(k) for k in x)
             for row,I in zip(rows,iters):
-                XZ[str(make_x(envs[row["environment_id"]]))].append(list(I)[-1])
+                XZ[str(make_x(row["environment_id"]))].append(list(I)[-1])
 
             if not is_scatter:
                 points = [(x,None)+YE(z) for x,z in XZ.items()]
@@ -698,6 +705,49 @@ class Result:
 
         return TransactionIO(filename).read()
 
+    @staticmethod
+    def from_logged_envs(environments: Sequence[Environment]):
+
+        environments = [e for e in environments if e.params.get('logged')]
+
+        envs_params = []
+        lrns_params = []
+
+        env_rows = {}
+        lrn_rows = {}
+        int_rows = {}
+
+        my_mean = lambda x: sum(x)/len(x)
+
+        for env in environments:
+            
+            is_batched = 'batched' in env.params
+            env_params  = dict(env.params)
+            lrn_params  = env_params.pop('learner')
+
+            try:
+                env_id = envs_params.index(env_params)
+            except:
+                env_id = len(envs_params)
+                envs_params.append(env_params)
+                env_rows[env_id] = env_params
+
+            try:
+                lrn_id = lrns_params.index(lrn_params)
+            except:
+                lrn_id = len(lrns_params)
+                lrns_params.append(lrn_params)
+                lrn_rows[lrn_id] = lrn_params
+
+            if is_batched:
+                results = {'_packed':{'reward': list(map(my_mean,map(itemgetter('reward'),env.read())))}}
+            else:
+                results = {'_packed':{'reward': list(map(itemgetter('reward'),env.read())) }}
+
+            int_rows[(env_id,lrn_id)]= results
+
+        return Result(env_rows,lrn_rows,int_rows)
+
     def __init__(self,
         env_rows: Dict[int           ,Dict[str,Any]] = {},
         lrn_rows: Dict[int           ,Dict[str,Any]] = {},
@@ -833,7 +883,7 @@ class Result:
         learner_id2: int,
         x          : Union[str, Sequence[str]] = "environment_id",
         y          : str = "reward",
-        mode       : Literal["diff","prob",'scat'] = "diff",
+        mode       : Union[Literal["diff","prob",'scat'],Callable[[float,float],float]] = "diff",
         span       : int = None,
         err        : Union[Literal['se','sd','bs'], None, PointAndInterval] = None,
         labels     : Sequence[str] = None,
@@ -842,6 +892,7 @@ class Result:
         ylim       : Tuple[Optional[Number],Optional[Number]] = None,
         xticks     : bool = True,
         yticks     : bool = True,
+        reverse    : bool = False,
         out        : Union[None,Literal['screen'],str] = 'screen',
         ax = None) -> None:
         """Plot a direct contrast of the performance for two learners.
@@ -864,68 +915,82 @@ class Result:
             ylim: Define the y-axis limits to plot. If `None` the y-axis limits will be inferred.
             xticks: Whether the x-axis labels should be drawn.
             yticks: Whether the y-axis labels should be drawn.
+            reverse: Whether to reverse order when counting winners and losers.
             out: Indicate where the plot should be sent to after plotting is finished.
             ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
         """
+        try:
+            xlim = xlim or [None,None]
+            ylim = ylim or [None,None]
 
-        xlim = xlim or [None,None]
-        ylim = ylim or [None,None]
+            x = [x] if isinstance(x,str) else list(x)
+            self._validate_parameters(x)
 
-        x = [x] if isinstance(x,str) else list(x)
-        self._validate_parameters(x)
+            if x == ['index']:
+                raise CobaException("plot_contrast does not currently support contrasting by `index`.")
 
-        rows = FilterPlottingData().filter(list(self.interactions), x, y, [learner_id1, learner_id2])
-        rows = SmoothPlottingData().filter(rows, y, span)
-        rows = ContrastPlottingData().filter(rows, y, mode, learner_id1)
+            rows = FilterPlottingData().filter(list(self.interactions), x, y, [learner_id1, learner_id2])
+            rows = SmoothPlottingData().filter(rows, y, span)
+            rows = ContrastPlottingData().filter(rows, y, mode, learner_id1)
 
-        XYE = TransformToXYE().filter(rows, self.environments, x, y, err)
+            XYE = TransformToXYE().filter(rows, self.environments, x, y, err)
 
-        if x != ['index']:
-            XYE = sorted(XYE, key=lambda xye: xye[1])
+            if x != ['index']:
+                XYE = sorted(XYE, key=lambda xye: xye[1], reverse=reverse)
 
-        bound = .5 if mode == "prob" else 0
+            bound = .5 if mode == "prob" else 0
 
-        win,tie,loss = [],[],[]
-        for _x,_y,_xe,_ye in XYE:
-            xl,xu = (0,0) if _xe is None else (_xe,_xe) if isinstance(_xe,Number) else _xe
-            yl,yu = (0,0) if _ye is None else (_ye,_ye) if isinstance(_ye,Number) else _ye
-            if mode != 'scat':
-                (win if _y-yl > bound else loss if _y+yu < bound else tie).append((_x,_y,_xe,_ye))
+            win,tie,loss = [],[],[]
+            for _x,_y,_xe,_ye in XYE:
+                o = 1 if not reverse else -1
+
+                xl,xu = (0,0) if _xe is None else (_xe,_xe) if isinstance(_xe,Number) else _xe
+                yl,yu = (0,0) if _ye is None else (_ye,_ye) if isinstance(_ye,Number) else _ye
+
+                if mode != 'scat':
+                    (win if _y-yl > o*bound else loss if _y+yu < o*bound else tie).append((_x,_y,_xe,_ye))
+                else:
+                    (win if _x-xl>o*(_y+yu) else loss if _y-yl> o*(_x+xu) else tie).append((_x,_y,_xe,_ye))
+
+            colors = (colors or []) + [0,1,2]
+            labels = (labels or []) + [self._full_name(learner_id1), self._full_name(learner_id2)]
+
+            fmt = "-" if x == ['index'] else "."
+
+            plots = []
+
+            if not reverse:
+                if loss: plots.append(Points(*zip(*loss), colors[0], 1, labels[1] + " " + f"({len(loss)})", fmt))
+                if tie : plots.append(Points(*zip(*tie) , colors[1], 1, 'Tie'     + " " + f"({len(tie )})", fmt))
+                if win : plots.append(Points(*zip(*win) , colors[2], 1, labels[0] + " " + f"({len(win )})", fmt))
             else:
-                (win if _x-xl>_y+yu else loss if _y-yl>_x+xu else tie).append((_x,_y,_xe,_ye))
+                if win : plots.append(Points(*zip(*win) , colors[2], 1, labels[0] + " " + f"({len(win )})", fmt))
+                if tie : plots.append(Points(*zip(*tie) , colors[1], 1, 'Tie'     + " " + f"({len(tie )})", fmt))
+                if loss: plots.append(Points(*zip(*loss), colors[0], 1, labels[1] + " " + f"({len(loss)})", fmt))
 
-        colors = (colors or []) + [0,1,2]
-        labels = (labels or []) + [self.learners[learner_id1]['full_name'], self.learners[learner_id2]['full_name']]
+            if mode != 'scat':
+                leftmost_x  = (loss+tie+win)[ 0][0] if not reverse else (win+tie+loss)[ 0][0]
+                rightmost_x = (loss+tie+win)[-1][0] if not reverse else (win+tie+loss)[-1][0]
+                plots.append(Points((leftmost_x,rightmost_x),(bound,bound), None, None , "#888", 1, None, '-',.5))
+            else:
+                m = max([p[0] for p in (loss+tie+win)]+[p[1] for p in (loss+tie+win)]+[1])
+                plots.append(Points((0,m),(0,m), None, None , "#888", 1, None, '-',.5))
 
-        fmt = "-" if x == ['index'] else "."
+            xrotation = 90 if x != ['index'] and len(XYE)>5 else 0
+            yrotation = 0
 
-        plots = []
+            if mode != "scat":
+                xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
+                ylabel = f"{labels[0]} - {labels[1]}" if mode=="diff" else f"P({labels[0]} > {labels[1]})"
+            else:
+                xlabel = y
+                ylabel = y
 
-        if loss: plots.append(Points(*zip(*loss), colors[0], 1, labels[1] + " " + f"({len(loss)})", fmt))
-        if tie : plots.append(Points(*zip(*tie) , colors[1], 1, 'Tie'     + " " + f"({len(tie )})", fmt))
-        if win : plots.append(Points(*zip(*win) , colors[2], 1, labels[0] + " " + f"({len(win )})", fmt))
+            title = f"{ylabel} ({len(rows) if x==['index'] else len(XYE)} Environments)"
 
-        if mode != 'scat':
-            leftmost_x  = (loss+tie+win)[ 0][0]
-            rightmost_x = (loss+tie+win)[-1][0]
-            plots.append(Points((leftmost_x,rightmost_x),(bound,bound), None, None , "#888", 1, None, '-'))
-        else:
-            m    = max([p[0] for p in (loss+tie+win)]+[p[1] for p in (loss+tie+win)]+[1])
-            plots.append(Points((0,m),(0,m), None, None , "#888", 1, None, '-'))
-
-        xrotation = 90 if x != ['index'] and len(XYE)>5 else 0
-        yrotation = 0
-
-        if mode != "scat":
-            xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
-            ylabel = f"{labels[0]} - {labels[1]}" if mode=="diff" else f"P({labels[0]} > {labels[1]})"
-        else:
-            xlabel = y
-            ylabel = y
-
-        title  = f"{ylabel} ({len(rows) if x==['index'] else len(XYE)} Environments)"
-
-        self._plotter.plot(ax, plots, title, xlabel, ylabel, xlim, ylim, xticks, yticks, xrotation, yrotation, out)
+            self._plotter.plot(ax, plots, title, xlabel, ylabel, xlim, ylim, xticks, yticks, xrotation, yrotation, out)
+        except CobaException as e:
+            CobaContext.logger.log(str(e)) 
 
     def plot_learners(self,
         ids   : Union[int,Sequence[int]] = None, 
@@ -934,7 +999,7 @@ class Result:
         span  : int = None,
         err   : Union[Literal['se','sd','bs'], None, PointAndInterval] = None,
         labels: Sequence[str] = None,
-        colors: Sequence[Union[str,int]] = None,
+        colors: Union[int,Sequence[Union[str,int]]] = None,
         xlim  : Tuple[Optional[Number],Optional[Number]] = None,
         ylim  : Tuple[Optional[Number],Optional[Number]] = None,
         xticks: bool = True,
@@ -961,53 +1026,83 @@ class Result:
             ylim: Define the y-axis limits to plot. If `None` the y-axis limits will be inferred.
             xticks: Whether the x-axis labels should be drawn.
             yticks: Whether the y-axis labels should be drawn.
-            top_n: Only plot the top_n learners. If `None` all learners will be plotted.
+            top_n: Only plot the top_n learners. If `None` all learners will be plotted. If negative the bottom will be plotted.
             out: Indicate where the plot should be sent to after plotting is finished.
             ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
         """
+        try:
+            xlim = xlim or [None,None]
+            ylim = ylim or [None,None]
 
-        xlim = xlim or [None,None]
-        ylim = ylim or [None,None]
+            if isinstance(ids,int): ids = [ids]
+            if isinstance(labels,str): labels = [labels]
+            if isinstance(x,str): x = [x]
 
-        if isinstance(ids,int): ids = [ids]
-        if isinstance(colors,int): colors = [colors]
-        if isinstance(labels,str): labels = [labels]
-        if isinstance(x,str): x = [x]
+            self._validate_parameters(x)
 
-        self._validate_parameters(x)
+            interactions = self.interactions if not ids else self.filter_lrn(learner_id=ids).interactions
 
-        interactions = self.interactions if not ids else self.filter_lrn(learner_id=ids).interactions
+            rows = FilterPlottingData().filter(list(interactions), x, y)
+            rows = SmoothPlottingData().filter(rows, y, span)
 
-        rows = FilterPlottingData().filter(list(interactions), x, y)
-        rows = SmoothPlottingData().filter(rows, y, span)
+            env_rows = self.environments
+            get_key  = lambda row: row['learner_id']
+            lines    = []
 
-        env_rows = self.environments
-        get_key  = lambda row: row['learner_id']
-        lines    = []
+            style = "-" if x == ['index'] else "."
 
-        style = "-" if x == ['index'] else "."
+            def get_color(colors:Union[None,Sequence[Union[str,int]]], i:int):
+                try:
+                    return colors[i] if colors else i
+                except IndexError:
+                    return i+max(colors) if isinstance(colors[0],(int,float)) else i
+                except TypeError:
+                    return i+colors
 
-        def get_color(colors:Sequence[Union[str,int]], i:int):
-            return i if not colors else i+max(colors) if isinstance(colors[0],int) else colors[i] if i < len(colors) else i
+            def get_label(labels:Sequence[str], i:int, lrn_id:int=None):
+                try:
+                    return labels[i] if labels else self._full_name(lrn_id)
+                except:
+                    return self._full_name(lrn_id)
 
-        for i, (lrn_id, lrn_rows) in enumerate(groupby(sorted(rows, key=get_key),key=get_key)):
-            lrn_rows = list(lrn_rows)
-            XYE      = TransformToXYE().filter(lrn_rows, env_rows, x, y, err)
-            color    = get_color(colors,i)
-            label    = labels[i] if labels and i < len(labels) else self.learners[lrn_id]['full_name']
-            lines.append(Points(*zip(*XYE), color, 1, label, style))
+            for i, (lrn_id, lrn_rows) in enumerate(groupby(sorted(rows, key=get_key),key=get_key)):
+                lrn_rows = list(lrn_rows)
+                XYE      = TransformToXYE().filter(lrn_rows, env_rows, x, y, err)
+                color    = get_color(colors,i)
+                label    = get_label(labels,i,lrn_id)
+                lines.append(Points(*zip(*XYE), color, 1, label, style))
 
-        lines  = sorted(lines, key=lambda line: line[1][-1], reverse=True)
-        xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
-        ylabel = y.capitalize().replace("_pct"," Percent")
+            lines  = sorted(lines, key=lambda line: line[1][-1], reverse=True)
+            labels = [l.label for l in lines]
+            colors = [l.color for l in lines]
+            xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
+            ylabel = y.capitalize().replace("_pct"," Percent")
 
-        title = ("Instantaneous" if span == 1 else f"Span {span}" if span else "Progressive") + f" {ylabel}"
-        title = title + f" ({len(lrn_rows) if x==['index'] else len(XYE)} Environments)"
+            title = ("Instantaneous" if span == 1 else f"Span {span}" if span else "Progressive") + f" {ylabel}"
+            title = title + f" ({len(lrn_rows) if x==['index'] else len(XYE)} Environments)"
 
-        if x != ['index']: title = f"Final {title}"
-        if top_n         : lines = [l._replace(color=get_color(colors,i)) for i,l in enumerate(lines[:top_n]) ]
+            if x != ['index']: title = f"Final {title}"
 
-        self._plotter.plot(ax, lines, title, xlabel, ylabel, xlim, ylim, xticks, yticks, 0, 0, out)
+            xrotation = 90 if x != ['index'] and len(XYE)>5 else 0
+            yrotation = 0
+
+            if top_n:
+                if abs(top_n) > len(lines): top_n = len(lines)*abs(top_n)/top_n
+                if top_n > 0: lines = [l._replace(color=get_color(colors,i),label=get_label(labels,i)) for i,l in enumerate(lines[:top_n],0    ) ]
+                if top_n < 0: lines = [l._replace(color=get_color(colors,i),label=get_label(labels,i)) for i,l in enumerate(lines[top_n:],top_n) ]
+
+            self._plotter.plot(ax, lines, title, xlabel, ylabel, xlim, ylim, xticks, yticks, xrotation, yrotation, out)
+        except CobaException as e:
+            CobaContext.logger.log(str(e)) 
+
+    def _full_name(self,lrn_id:int) -> str:
+        """A user-friendly name created from a learner's params for reporting purposes."""
+
+        values = self.learners[lrn_id]
+        family = values.get('family',values['learner_id'])
+        params = f"({','.join(f'{k}={v}' for k,v in values.items() if k not in ['family','learner_id'])})"
+
+        return family if params == '()' else family+params
 
     def __str__(self) -> str:
         return str({"Learners": len(self._learners), "Environments": len(self._environments), "Interactions": len(self._interactions) })
