@@ -1,23 +1,24 @@
-import time
 import collections
 import collections.abc
-import math
+import time
 import warnings
-
 from abc import ABC, abstractmethod
-from statistics import mean
 from itertools import combinations
-from typing import Iterable, Any, Sequence, Mapping, Hashable, Union, overload
-from coba.backports import Literal
+from statistics import mean
+from typing import Iterable, Any, Sequence, Mapping, Hashable
 
-from coba.statistics import percentile
+import math
+
+from coba.backports import Literal
+from coba.contexts import CobaContext
+from coba.encodings import InteractionsEncoder
+from coba.environments import Environment, SafeEnvironment, Interaction
 from coba.exceptions import CobaExit
 from coba.learners import Learner, SafeLearner
-from coba.encodings import InteractionsEncoder
-from coba.utilities import PackageChecker, peek_first
-from coba.contexts import CobaContext
 from coba.primitives import Batch
-from coba.environments import Environment, SafeEnvironment, Interaction
+from coba.statistics import percentile
+from coba.utilities import PackageChecker, peek_first
+
 
 class LearnerTask(ABC):
     """A task which describes a Learner."""
@@ -68,62 +69,36 @@ class EvaluationTask(ABC):
 
 class SimpleEvaluation(EvaluationTask):
 
-    @overload
-    def __init__(self,
-        reward_metrics: Union[str,Sequence[str]] = "reward",
-        time_metrics: bool = False,
-        probability: bool = False) -> None:
-        ...
-        """
-        Args:
-            reward_metrics: The reward metrics to that should be recorded for each example.
-                The metric options include: reward, rank, regret.
-            time_metrics: Indicates whether time metrics should be recorded.
-            probability: Indicates whether the action's probability should be recorded.
-        """
-
-    @overload
-    def __init__(self, record: Sequence[Literal['reward','rank','regret','time','prob']] = ['reward']) -> None:
-        ...
+    def __init__(self, 
+        record: Sequence[Literal['reward','rank','regret','time','probability','action', 'context', 'ope_loss']] = ['reward'],
+        learn: bool = True,
+        predict: bool = True) -> None:
         """
         Args:
             record: The datapoints to record for each interaction.
+            learn: Indicates if learning should occur during the evaluation process.
+            predict: Indicates if predictions should occur during the evaluation process.
         """
 
-    def __init__(self, *args, **kwargs) -> None:
+        self._record  = [record] if isinstance(record,str) else record
+        self._learn   = learn
+        self._predict = predict
 
-        is_old_interface = (len(args) > 1 and isinstance(args[1],bool)) or (len(kwargs) > 0 and 'record' not in kwargs)
-
-        if not is_old_interface:
-            if not args and not kwargs: args = [['reward']]
-            record = args[0] if args else kwargs['record']
-            if isinstance(record,str): record = [record]
-        else:
-            record = []
-
-            if 'reward_metrics' in kwargs:
-                args = [kwargs['reward_metrics']]
-            if 'time_metrics' in kwargs:
-                args = ["reward"] if not args else list(args)
-                args.append(kwargs['time_metrics'])
-            if 'probability' in kwargs:
-                if len(args) == 1: args = list(args) + [False]
-                if len(args) == 0: args = ["reward", False]
-                args.append(kwargs['probability']) 
-
-            if len(args) > 0 and     isinstance(args[0],str): record.append(args[0])
-            if len(args) > 0 and not isinstance(args[0],str): record.extend(args[0])
-            if len(args) > 1 and args[1]                    : record.append("time")
-            if len(args) > 2 and args[2]                    : record.append("prob")
-
-        self._record = record
+        if 'ope_loss' in self._record:
+            # OPE loss metric is only available for VW models
+            # Divide by the number of samples for the average loss metric and see this article for more info
+            # https://vowpalwabbit.org/docs/vowpal_wabbit/python/latest/tutorials/off_policy_evaluation.html
+            PackageChecker.vowpalwabbit('SimpleEvaluation.__init__')
 
     def process(self, learner: Learner, interactions: Iterable[Interaction]) -> Iterable[Mapping[Any,Any]]:
 
         learner = SafeLearner(learner)
 
-        record_prob = 'prob' in self._record
-        record_time = 'time' in self._record
+        record_prob    = 'probability' in self._record
+        record_time    = 'time'        in self._record
+        record_action  = 'action'      in self._record
+        record_context = 'context'     in self._record
+        record_ope_loss = 'ope_loss'     in self._record
 
         first, interactions = peek_first(interactions)
 
@@ -147,33 +122,48 @@ class SimpleEvaluation(EvaluationTask):
 
         learning_info.clear()
 
+        should_pred  = self._predict
+        should_learn = self._learn
+
+        def indexify(action,actions,discrete,batched):
+            if not discrete: return action
+            return [ A.index(a) for a,A in zip(action,actions) ] if batched else actions.index(action)
+
         for interaction in interactions:
 
             interaction = interaction.copy()
 
+            is_on_policy_eval   = 'actions' in interaction and 'rewards' in interaction
+            is_off_policy_learn = 'action'  in interaction and ('reward' in interaction or 'feedback' in interaction)
+            is_on_policy_learn  = is_on_policy_eval and not is_off_policy_learn
+
+            is_predict = should_pred and is_on_policy_eval
+            is_learn   = should_learn and (is_off_policy_learn or is_on_policy_learn)
+
             out     = {}
             context = interaction.pop('context')
             batched = isinstance(context, Batch)
+            discrete = len(interaction.get('actions',[])) > 0
 
-            is_on_policy_eval   = 'actions' in interaction and 'rewards' in interaction
-            is_off_policy_learn = 'action' in interaction and 'reward' in interaction
-            is_on_policy_learn  = is_on_policy_eval and not is_off_policy_learn
-
-            is_predict = is_on_policy_eval
-            is_learn   = is_off_policy_learn or is_on_policy_learn
+            if record_context: out['context'] = context
 
             if is_predict:
-                actions = interaction.pop('actions')
-                rewards = interaction.pop('rewards')
+                actions   = interaction.pop('actions')
+                rewards   = interaction.pop('rewards')
+                feedbacks = interaction.pop('feedbacks',None)
 
                 start_time       = time.time()
                 action,prob,info = predict(context, actions)
                 predict_time     = time.time()-start_time
 
-                reward = rewards.eval(action)
+                _action  = indexify(action,actions,discrete,batched)
+                reward   = rewards.eval(_action)
+                feedback = feedbacks.eval(_action) if feedbacks else reward
 
-                if record_time: out['predict_time'] = predict_time
-                if record_prob: out['probability']  = prob
+                if record_time  : out['predict_time'] = predict_time
+                if record_prob  : out['probability']  = prob
+                if record_action: out['action']       = action
+                if feedbacks    : out['feedback']     = feedback
 
                 if not batched:
                     if calc_reward : out['reward'] = get_reward(reward)
@@ -184,26 +174,33 @@ class SimpleEvaluation(EvaluationTask):
                     if calc_regret : out['regret'] = mean(map(get_regret,reward,rewards))
                     if calc_rank   : out['rank'  ] = mean(map(get_rank  ,reward,rewards,len(actions)))
 
-                if 'feedbacks' in interaction:
-                    reward = interaction.pop('feedbacks').eval(action)
-                    out['feedback'] = reward
-
             if is_learn:
-
                 if is_off_policy_learn:
-                    actions = interaction.pop('actions',actions if is_predict else None)
-                    action  = interaction.pop('action')
-                    reward  = interaction.pop('reward')
-                    prob    = interaction.pop('probability',None)
-                    info    = {}
+                    actions  = interaction.pop('actions',actions if is_predict else None)
+                    action   = interaction.pop('action')
+                    reward   = interaction.pop('reward')
+                    feedback = reward if 'feedback' not in interaction else interaction.pop('feedback')
+                    prob     = interaction.pop('probability',None)
+                    info     = {}
 
                 start_time = time.time()
-                learn(context, actions, action, reward, prob, **info)
+                learn(context, actions, action, feedback, prob, **info)
                 learn_time = time.time() - start_time
 
                 if record_time: out['learn_time'] = learn_time
+                if record_ope_loss:
+                    # OPE loss metric is only available for VW models
+                    try:
+                        out['ope_loss'] = learner._learner._vw._vw.get_sum_loss()
+                    except AttributeError:
+                        out['ope_loss'] = float("nan")
 
             if interaction:
+                interaction.pop("actions",None)
+                interaction.pop("rewards",None)
+                interaction.pop("action",None)
+                interaction.pop("reward",None)
+                interaction.pop("probability",None)
                 out.update(interaction)
 
             if learning_info:
@@ -216,8 +213,7 @@ class SimpleLearnerInfo(LearnerTask):
     """Describe a Learner using its name and hyperparameters."""
 
     def process(self, item: Learner) -> Mapping[Any,Any]:
-        item = SafeLearner(item)
-        return {"full_name": item.full_name, **item.params}
+        return SafeLearner(item).params
 
 class SimpleEnvironmentInfo(EnvironmentTask):
     """Describe an Environment using its Environment and Filter parameters."""
