@@ -3,10 +3,9 @@ from typing import Any, Dict, Sequence
 from coba.exceptions import CobaException
 from coba.utilities import PackageChecker
 from coba.primitives import Context, Action
-from coba.pipes import Flatten
 from coba.encodings import InteractionsEncoder
 
-from coba.learners.primitives import Probs, Actions, Learner
+from coba.learners.primitives import PMF, Actions, Learner
 
 class LinUCBLearner(Learner):
     """A contextual bandit learner that represents expected reward as a
@@ -56,7 +55,7 @@ class LinUCBLearner(Learner):
     def params(self) -> Dict[str, Any]:
         return {'family': 'LinUCB', 'alpha': self._alpha, 'features': self._X}
 
-    def predict(self, context: Context, actions: Actions) -> Probs:
+    def predict(self, context: Context, actions: Actions) -> PMF:
 
         import numpy as np #type: ignore
 
@@ -66,7 +65,7 @@ class LinUCBLearner(Learner):
         if not context:
             self._X_encoder = InteractionsEncoder(list(set(filter(None,[ f.replace('x','') if isinstance(f,str) else f for f in self._X ]))))
 
-        context = list(Flatten().filter([list(context)]))[0] if context else []
+        context = context or []
         features: np.ndarray = np.array([self._X_encoder.encode(x=context,a=action) for action in actions]).T
 
         if(self._A_inv is None):
@@ -79,21 +78,19 @@ class LinUCBLearner(Learner):
         action_values = point_estimate + self._alpha*np.sqrt(point_bounds)
         max_indexes   = np.where(action_values == np.amax(action_values))[0]
 
-        return Probs([int(ind in max_indexes)/len(max_indexes) for ind in range(len(actions))])
+        return PMF([int(ind in max_indexes)/len(max_indexes) for ind in range(len(actions))])
 
     def learn(self, context: Context, actions: Actions, action: Action, reward: float, probability: float) -> None:
 
         import numpy as np
 
-        action = actions[0]
-
         if isinstance(action, dict) or isinstance(context, dict):
             raise CobaException("Sparse data cannot be handled by this algorithm.")
 
         if not context:
             self._X_encoder = InteractionsEncoder(list(set(filter(None,[ f.replace('x','') if isinstance(f,str) else f for f in self._X ]))))
 
-        context = list(Flatten().filter([list(context)]))[0] if context else []
+        context = context or []
         features: np.ndarray = np.array(self._X_encoder.encode(x=context,a=action)).T
 
         if(self._A_inv is None):
