@@ -25,5 +25,5 @@ __all__ = [
     'AIndex',
     'Batch',
     'BatchReward',
-    'BatchFeedback'
+    'BatchFeedback',
 ]
\ No newline at end of file
